from __future__ import annotations

# Standard library imports.
from typing import Optional

# Project imports.
import disruptive.log as dtlog
import disruptive.requests as dtrequests
import disruptive.events.events as dtevents
import disruptive.outputs as dtoutputs


class Device(dtoutputs.OutputBase):
    """
    Represents sensors and cloud connectors.

    When a device response is received, the content is
    unpacked and the related attributes are updated.

    Attributes
    ----------
    device_id : str
        Unique device ID.
    project_id : str
        Project in which the device resides.
    display_name : str, None
        Given device name if set through a label with key `name`.
        Otherwise None.
    type : str
        Device type.
    labels : dict
        Label keys and values.
    is_emulator : bool
        True if the device is an emulator, otherwise False.
    reported : Reported, None
        Object containing the data from the most recent events.

    """

    def __init__(self, device: dict) -> None:
        """
        Constructs the Device object by unpacking the raw device response.

        Parameters
        ----------
        device : dict
            Unmodified device response dictionary.

        """

        # Inherit from OutputBase parent.
        dtoutputs.OutputBase.__init__(self, device)

        # Unpack attributes from dictionary.
        self.device_id = device['name'].split('/')[-1]
        self.project_id = device['name'].split('/')[1]
        self.type = device['type']
        self.labels = device['labels']

        # Set display_name if `name` label key exists.
        self.display_name = None
        if 'name' in self.labels:
            self.display_name = self.labels['name']

        # Determine if the device is an emulator by checking id prefix.
        if self.device_id.startswith('emu'):
            self.is_emulator = True
        else:
            self.is_emulator = False

        # If it exists, set the reported object.
        self.reported = None
        if 'reported' in device:
            self.reported = Reported(device['reported'])

    @classmethod
    def get_device(cls,
                   device_id: str,
                   project_id: Optional[str] = None,
                   **kwargs,
                   ) -> Device:
        """
        Gets a device specified by its project and ID.

        Parameters
        ----------
        device_id : str
            Unique ID of the target device.
        project_id : str, optional
            Unique ID of the target project.
            If this is not provided, a wildcard project will be used, resulting
            in a search for the device through all available projects.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        device : Device
            Object representing the specified device.

        Examples
        --------
        >>> device = dt.Device.get_device(project_id, device_id)

        """

        if project_id is None:
            project_id = '-'

        # Construct URL
        url = '/projects/{}/devices/{}'.format(project_id, device_id)

        # Return Device object of GET request response.
        return cls(dtrequests.DTRequest.get(url, **kwargs))

    @classmethod
    def list_devices(cls,
                     project_id: str,
                     query: Optional[str] = None,
                     device_ids: Optional[list[str]] = None,
                     device_types: Optional[list[str]] = None,
                     label_filters: Optional[dict[str, str]] = None,
                     order_by: Optional[str] = None,
                     **kwargs,
                     ) -> list[Device]:
        """
        List all available devices in the specified project.

        Parameters
        ----------
        project_id : str
            Unique ID of the target project.
        query : str, optional
            Keyword based search for device display name.
        device_ids : list[str], optional
            Specify devices by their unique IDs.
        device_types : list[str], optional
            Specify devices by type.
        label_filters : dict[str, str], optional
            Specify devices by label keys and values (i.e. {"key": "value"}).
            If a key is provided and value is set as an empty
            string (i.e. {"key": ""}), all device with that key is fetched.
        order_by : str, optional
            The field name you want to order the response by.
            Referred to using dot notation, e.g. reported.temperature.value.
            Default order is ascending, but can be flipped by prefixing a "~".
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        devices : list[Device]
            List of objects each representing a device.

        """

        # Construct parameters dictionary.
        params: dict = dict()
        if query is not None:
            params['query'] = query
        if device_ids is not None:
            params['device_ids'] = device_ids
        if device_types is not None:
            params['device_types'] = device_types
        if order_by is not None:
            params['order_by'] = order_by

        # Convert label_filters dictionary to list of strings.
        if label_filters is not None:
            labels_list = []
            for key in label_filters:
                labels_list.append(key + '=' + label_filters[key])
            params['label_filters'] = labels_list

        # Return list of Device objects of paginated GET response.
        devices = dtrequests.DTRequest.paginated_get(
            url='/projects/{}/devices'.format(project_id),
            pagination_key='devices',
            params=params,
            **kwargs,
        )
        return [cls(device) for device in devices]

    @staticmethod
    def batch_update_labels(device_ids: list[str],
                            project_id: str,
                            set_labels: Optional[dict[str, str]] = None,
                            remove_labels: Optional[list[str]] = None,
                            **kwargs,
                            ) -> None:
        """
        Add, update, or remove multiple labels (key=value) on multiple devices

        Must provide either add_labels or remove_labels. If neither are
        provided, a BadRequest error will be raised.

        If a key provided with `add_labels` already exists for a
        device, the value will be updated accordingly.

        Parameters
        ----------
        device_ids : list[str]
            List of unique IDs for the target devices.
        project_id : str
            Unique ID of target project.
        add_labels : dict[str, str], optional
            Key and value of labels to be added / updated.
        remove_labels : list[str], optional
            Label keys to be removed.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        """

        # Construct list of devices.
        name = 'projects/{}/devices/{}'
        devices = [name.format(project_id, xid) for xid in device_ids]

        # Construct request body dictionary.
        body: dict = dict()
        body['devices'] = devices
        if set_labels is not None:
            body['addLabels'] = set_labels
        if remove_labels is not None:
            body['removeLabels'] = remove_labels

        # Construct URL.
        url = '/projects/{}/devices:batchUpdate'.format(project_id)

        # Sent POST request, but return nothing.
        dtrequests.DTRequest.post(url, body=body, **kwargs)

    @staticmethod
    def set_label(device_id: str,
                  project_id: str,
                  key: str,
                  value: str,
                  **kwargs,
                  ) -> None:
        """
        Add a label (key=value) for a single device.
        If key already exists, the value is updated.

        Parameters
        ----------
        device_id : str
            Unique ID of the target device.
        project_id : str
            Unique ID of the target project.
        key : str
            Label key to be added.
        value : str
            Label value to be added.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        """

        # Construct URL.
        url = '/projects/{}/devices:batchUpdate'.format(project_id)

        # Construct request body dictionary.
        body: dict = dict()
        body['devices'] = ['projects/' + project_id + '/devices/' + device_id]
        body['addLabels'] = {key: value}

        # Sent POST request, but return nothing.
        dtrequests.DTRequest.post(url, body=body, **kwargs)

    @staticmethod
    def remove_label(device_id: str,
                     project_id: str,
                     key: str,
                     **kwargs,
                     ) -> None:
        """
        Remove a label (key=value) from a single device.

        Parameters
        ----------
        device_id : str
            Unique ID of the target device.
        project_id : str
            Unique ID of the target project.
        key : str
            Key of the label to be removed.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        """

        # Construct URL.
        url = '/projects/{}/devices:batchUpdate'.format(project_id)

        # Construct request body dictionary.
        body: dict = dict()
        body['devices'] = ['projects/' + project_id + '/devices/' + device_id]
        body['removeLabels'] = [key]

        # Sent POST request, but return nothing.
        dtrequests.DTRequest.post(url, body=body, **kwargs)

    @staticmethod
    def transfer_devices(device_ids: list[str],
                         source_project_id: str,
                         target_project_id: str,
                         **kwargs,
                         ) -> None:
        """
        Transfer devices from one project to another.

        Parameters
        ----------
        device_ids : list[str]
            List of unique IDs for the target devices.
        source_project_id : str
            Unique ID of the source project.
        target_project_id : str
            Unique ID of the target project.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        """

        # Construct list of devices.
        name = 'projects/{}/devices/{}'
        devices = [name.format(source_project_id, xid) for xid in device_ids]

        # Construct request body dictionary.
        body = {
            "devices": devices
        }

        # Sent POST request, but return nothing.
        dtrequests.DTRequest.post(
            url='/projects/{}/devices:transfer'.format(
                target_project_id
            ),
            body=body,
            **kwargs,
        )


class Reported(dtoutputs.OutputBase):
    """
    Represents the "reported" field for a device.

    Contains one attribute for each event type, initialized to None.
    For each event type represented in the reported field,
    the related attribute is updated with the
    appropriate :ref:`Event Data <Event Data>` class.

    Attributes
    ----------
    touch : Touch, None
        Object representing reported touch event data.
    temperature : Temperature, None
        Object representing reported temperature event data.
    object_present : ObjectPresent, None
        Object representing reported objectPresent event data.
    humidity : Humidity, None
        Object representing reported humidity event data.
    object_present_count : ObjectPresentCount, None
        Object representing reported objectPresentCount event data.
    touch_count : TouchCount, None
        Object representing reported touchCount event data.
    water_present : WaterPresent, None
        Object representing reported waterPresent event data.
    network_status : NetworkStatus, None
        Object representing reported networkStatus event data.
    battery_status : BatteryStatus, None
        Object representing reported batteryStatus event data.
    connection_status : ConnectionStatus, None
        Object representing reported connectionStatus event data.
    ethernet_status : EthernetStatus, None
        Object representing reported ethernetStatus event data.
    cellular_status : CellularStatus, None
        Object representing reported cellularStatus event data.

    """

    def __init__(self, reported: dict) -> None:
        """
        Constructs the Reported object by unpacking each event in the field.

        Parameters
        ----------
        reported : dict
            Dictionary of the reported field for a device.

        """

        # Inherit parent Event class init.
        dtoutputs.OutputBase.__init__(self, reported)

        # Set default attribute values.
        for key in dtevents._EVENTS_MAP._api_names:
            # Skip labelsChanged as it does not exist in reported.
            if key == 'labelsChanged':
                continue

            # Set attribute to None.
            setattr(self, dtevents._EVENTS_MAP._api_names[key].attr_name, None)

        # Unpack the reported dictionary data.
        self.__unpack()

    def __unpack(self) -> None:
        """
        Iterates each field in the raw dictionary and set the
        related attribute to the appropriate _EventData child object.
        If an event type is not found, the attribute is left as None.

        """

        # Iterate keys in reported dictionary.
        for key in self._raw.keys():
            # Fields can be None on emulated devices. Skip if that is the case.
            if self._raw[key] is None:
                continue

            # Also skip labelsChanged key as it does not exist in reported.
            if key == 'labelsChanged':
                continue

            # Repack the data field in expected format.
            repacked = {key: self._raw[key]}

            # Initialize appropriate data instance.
            data = dtevents._EventData.from_event_type(repacked, key)

            # Set attribute according to event type.
            if key in dtevents._EVENTS_MAP._api_names:
                setattr(
                    self,
                    dtevents._EVENTS_MAP._api_names[key].attr_name,
                    data,
                )
            else:
                dtlog.log('Skipping unknown event type {}.'.format(key))



# Project imports.
import disruptive.transforms as dttrans


class OutputBase():
    """
    Represents common features for all returnable objects.

    """

    def __init__(self, raw: dict) -> None:
        """
        Constructs the OutputBase object by setting raw attribute.

        Parameters
        ----------
        raw : dict
            Unmodified dictionary of data received from the REST API.

        """

        # Set attribute from input argument.
        self._raw = raw

    def __str__(self):
        out = self.__dump([], self, level=0)
        return '\n'.join(out)

    def __dump(self, out, obj, level):
        # Set the two formatting indent levels.
        n_spaces = 4
        ls = level*' '*n_spaces

        # Print name of the current object if level is 0.
        if level == 0:
            out.append(ls + str(self.__class__.__name__) + ' object:')
            # out.insert(-1, ls + '_'*len(out[-1]))
            out.insert(-1, '')
            out.append(ls + '-'*len(out[-1]))

        # Append the various public attributes recursively.
        for a in vars(obj):
            # Skip private attributes.
            if a.startswith('_'):
                continue

            # Fetch and evaluate the attribute value / type.
            val = getattr(obj, a)
            if hasattr(val, '__dict__'):
                # Class objects should be dumped recursively.
                out.append('{}[{}] {}:'.format(ls, type(val).__name__, a))
                self.__dump(out, val, level=level+1)

            elif isinstance(val, list):
                # Lists content should be iterated through.
                out.append('{}[{}] {}:'.format(ls, type(val).__name__, a))
                self.__dump_list(out, val, level+1)

            else:
                # Other types can be printed directly.
                out.append('{}[{}] {}: {}'.format(
                    ls, type(val).__name__, a, str(val)
                ))

        return out

    def __dump_list(self, out, lst, level):
        n_spaces = 4
        ls = (level+1)*' '*n_spaces
        for val in lst:
            if hasattr(val, '__dict__'):
                out.append('{}[{}]:'.format(ls, type(val).__name__))
                self.__dump(out, val, level=level+2)
            else:
                out.append('{}[{}] {}'.format(
                    ls, type(val).__name__, val
                ))
        return out


class Metric(OutputBase):
    """
    Represents the metrics for a dataconnector over the last 3 hours.

    Attributes
    ----------
    success_count : int
        Number of 2xx responses.
    error_count : int
        Number of non-2xx responses.
    latency : str
        Average latency.

    """

    def __init__(self, metric: dict) -> None:
        """
        Constructs the Metric object by unpacking the raw response.

        """

        # Inherit attributes from ResponseBase parent.
        OutputBase.__init__(self, metric)

        # Unpack attributes from dictionary.
        self.success_count = metric['metrics']['successCount']
        self.error_count = metric['metrics']['errorCount']
        self.latency = metric['metrics']['latency99p']


class Member(OutputBase):
    """
    Represents a member.

    Attributes
    ----------
    display_name : str
        Provided member display name.
    roles : list[str]
        Roles provided to the member.
    status : str
        Whether the member invite is ACCEPTED or PENDING.
    email : str
        Member email address.
    account_type : str
        Whether the member is a USER or SERVICE_ACCOUNT.
    create_time : datetime
        Timestamp of when the member was created.

    """

    def __init__(self, member):
        """
        Constructs the Member object by unpacking the raw response.

        """

        # Inherit from Response parent.
        OutputBase.__init__(self, member)

        # Unpack attributes from dictionary.
        self.display_name = member['displayName']
        self.roles = [r.split('/')[-1] for r in member['roles']]
        self.status = member['status']
        self.email = member['email']
        self.account_type = member['accountType']
        self.create_time = dttrans.to_datetime(member['createTime'])





class DTApiError(Exception):
    """
    Represents errors raised from the REST API.

    """

    def __init__(self, message):

        # Call the base class constructor.
        super().__init__(message)


class BadRequest(DTApiError):
    """
    The response contained a status code of 400.
    https://developer.d21s.com/docs/error-codes#400

    """

    def __init__(self, message):
        super().__init__(message)


class Unauthorized(DTApiError):
    """
    The response contained a status code of 401.
    https://developer.d21s.com/docs/error-codes#401

    """

    def __init__(self, message):
        super().__init__(message)


class Forbidden(DTApiError):
    """
    The response contained a status code of 403.
    https://developer.d21s.com/docs/error-codes#403

    """

    def __init__(self, message):
        super().__init__(message)


class NotFound(DTApiError):
    """
    The response contained a status code of 404.
    https://developer.d21s.com/docs/error-codes#404

    """

    def __init__(self, message):
        super().__init__(message)


class Conflict(DTApiError):
    """
    The response contained a status code of 409.
    https://developer.d21s.com/docs/error-codes#409

    """

    def __init__(self, message):
        super().__init__(message)


class TooManyRequests(DTApiError):
    """
    The response contained a status code of 429.
    https://developer.d21s.com/docs/error-codes#429

    """

    def __init__(self, message):
        super().__init__(message)


class InternalServerError(DTApiError):
    """
    The response contained a status code of 500.
    https://developer.d21s.com/docs/error-codes#500

    """

    def __init__(self, message):
        super().__init__(message)


class ReadTimeout(DTApiError):
    """
    The server did not respond in the alloted amount
    of time set by :ref:`request_timeout <config params>`.

    """

    def __init__(self, message):
        super().__init__(message)


class ConnectionError(DTApiError):
    """
    Could not establish connection to the server.

    """

    def __init__(self, message):
        super().__init__(message)


class FormatError(DTApiError):
    """
    Some string formatting, usually a timestamp, is wrong.

    """

    def __init__(self, message):
        super().__init__(message)


class ConfigurationError(DTApiError):
    """
    One or more :ref:`configuration parameters <configuration>` are invalid.

    """

    def __init__(self, message):
        super().__init__(message)


class UnknownError(DTApiError):
    """
    An unexpected error has been raised.

    This is likely due to a bug in the package.
    Please report this to developer-support@disruptive-technologies.com.

    """

    def __init__(self, message):
        super().__init__(message)



# Standard library imports
from datetime import datetime
from typing import Optional

# Project imports.
import disruptive as dt


def log(msg: str, override: Optional[bool] = None) -> None:
    """
    Prints message to stdout if global flag is True.

    Parameters
    ----------
    msg : str
        Message to be printed.

    """

    # Initialize flag.
    should_log = False

    # First set flag from package-wide configuration.
    if dt.log is True:
        should_log = True

    # Override if provided.
    if override is not None:
        should_log = override

    if should_log:
        print('[{}]: {}'.format(
            datetime.now(),
            msg,
        ))



# Metadata
__version__ = "0.2.1"

# Authentication scheme.
from disruptive.authentication import Auth  # noqa

# When auth is an instance of Auth set with the default constructor,
# the package is unauthenticated, throwing an Unauthorized error at call.
# To authorize, either overwrite the global variable `auth` or provide
# each individual method with the same auth object.
#
# Available authentication methods:
# auth = disruptive.Auth.serviceaccount(key_id, secret, email)
#
# Initialize package with unauthenticated object.
default_auth = Auth({})

# Resources.
from disruptive.resources.device import Device  # noqa
from disruptive.resources.dataconnector import DataConnector  # noqa
from disruptive.resources.organization import Organization  # noqa
from disruptive.resources.project import Project  # noqa
from disruptive.resources.stream import Stream  # noqa
from disruptive.resources.eventhistory import EventHistory  # noqa
from disruptive.resources.serviceaccount import ServiceAccount  # noqa
from disruptive.resources.role import Role  # noqa
from disruptive.resources.emulator import Emulator  # noqa

# Additional helper modules.
import disruptive.types as types  # noqa
import disruptive.events as events  # noqa
import disruptive.dataconnector_configs as dataconnector_configs  # noqa

# If True, debug information is sent to stdout.
log = False

# REST API base URLs of which all endpoints are an expansion.
api_url = 'https://api.disruptive-technologies.com/v2'
emulator_url = 'https://emulator.disruptive-technologies.com/v2'
auth_url = 'https://identity.disruptive-technologies.com/oauth2/token'

# If a request response contains an error for which a series of retries is
# worth considering, these variable determine how long to wait without an
# answer, and how many times the package should retry before raising an error.
request_timeout = 3  # seconds
request_retries = 3  # attempts



from __future__ import annotations

# Standard library imports.
import time
import urllib.parse

# Third-party imports.
import jwt

# Project imports
import disruptive
import disruptive.requests as dtrequests
import disruptive.errors as dterrors


class Auth():
    """
    Used to initialize and maintain authentication to the REST API.

    Attributes
    ----------
    token : str
        Access token provided in request header.
    method : str
        Authentication method used.
    expiration : int
        Unixtime of when the authentication expires.
    credentials : dict[str, str]
        Credentials provided for authenticating with the chosen method.

    """

    def __init__(self, credentials: dict[str, str]) -> None:
        """
        Constructs the Auth object by validating and updating credentials.

        Parameters
        ----------
        credentials : dict[str, str]
            Credentials provided for authenticating with the chosen method.

        """

        # Verify provided credentials are strings.
        self._verify_str_credentials(credentials)

        # Initialize attributes with nonset values.
        self.token = ''
        self.expiration = 0

        # Set arguments as attributes
        self.credentials = credentials

    @classmethod
    def serviceaccount(cls,
                       key_id: str,
                       secret: str,
                       email: str,
                       ) -> Auth:
        """
        Constructs the Auth object for authenticating using a serviceaccount.

        This method uses an OAuth2 flow to authenticate. Using the provided
        credentials, a JWT is created and exchanged for an access token which
        is renewed every hour as required.

        Parameters
        ----------
        key_id : str
            Unique serviceaccount key ID.
        secret : str
            Serviceaccount secret.
        email : str
            Unique serviceaccount email address.

        Returns
        -------
        auth : Auth
            Object to initialize and maintain authentication to the REST API.

        """

        # Construct Auth object with method and credentials.
        obj = cls(
            credentials={
                'key_id': key_id,
                'secret': secret,
                'email': email,
            },
        )

        # Patch the newly created object with method-specific methods.
        setattr(obj, '_has_expired', obj.__serviceaccount_has_expired)
        setattr(obj, 'refresh', obj.__serviceaccount_refresh)

        # Return the patch object.
        return obj

    def _verify_str_credentials(self, credentials: dict) -> None:
        """
        Verifies that the provided credentials are strings.

        This check is added as people use environment variables, but
        if for instance os.environ.get() does not find one, it silently
        returns None. It's better to just check for it early.

        Parameters
        ----------
        credentials : dict
            Credentials used to authenticate the REST API.

        """

        for key in credentials:
            if type(credentials[key]) != str:
                raise TypeError(
                    'Authentication credentials must be of type '
                    + '<class \'str\'>, but received {}.'.format(
                        type(credentials[key])
                    ))

    def get_token(self) -> str:
        """
        Returns the access token.
        If the token has expired, renew it.

        Returns
        -------
        token : str
            Access token added to the request header.

        """

        # Check expiration time.
        if self._has_expired():
            # Renew access token.
            self.refresh()

        return self.token

    def _has_expired(self) -> bool:
        """
        Raises an Unauthorized error with some added information.
        If this function is called, the project has not yet been
        authenticated, and the user should be reminded to do so.

        """
        raise dterrors.Unauthorized(
            'No authentication method has been set.\n'
            'Package-wide authentication can be set by:'
            + '\n>>> import disruptive as dt'
            + '\n>>> dt.default_auth = dt.Auth.serviceaccount'
            + '(key_id, secret, email)'
        )

    def refresh(self) -> None:
        """
        This function does nothing until an authenticate method has
        been initialized. Until then, it acts as a name placeholder
        which is replaced by a type-specific method.
        """
        pass

    def __serviceaccount_has_expired(self) -> bool:
        """
        Evaluates whether the access token has expired.

        Returns
        -------
        has_expired : bool
            True if the access token has expired, otherwise False.

        """

        if time.time() > self.expiration:
            return True
        else:
            return False

    def __serviceaccount_refresh(self) -> None:
        """
        Refreshes the access token.

        This first exchanges the JWT for an access token, then updates
        the expiration and token attributes with the response.

        """

        response = self.__serviceaccount_get_access_token()
        self.expiration = time.time() + response['expires_in']
        self.token = 'Bearer {}'.format(response['access_token'])

    def __serviceaccount_get_access_token(self) -> dict:
        """
        Constructs and exchanges the JWT for an access token.

        Returns
        -------
        response : dict
            Dictionary containing expiration and the token itself.

        """

        # Set access token URL.
        token_url = disruptive.auth_url

        # Construct the JWT header.
        jwt_headers = {
            'alg': 'HS256',
            'kid': self.credentials['key_id'],
        }

        # Construct the JWT payload.
        jwt_payload = {
            'iat': int(time.time()),         # current unixtime
            'exp': int(time.time()) + 3600,  # expiration unixtime
            'aud': token_url,
            'iss': self.credentials['email'],
        }

        # Sign and encode JWT with the secret.
        encoded_jwt = jwt.encode(
            payload=jwt_payload,
            key=self.credentials['secret'],
            algorithm='HS256',
            headers=jwt_headers,
        )

        # Prepare HTTP POST request data.
        # Note: The requests package applies Form URL-Encoding by default.
        request_data = urllib.parse.urlencode({
            'assertion': encoded_jwt,
            'grant_type': 'urn:ietf:params:oauth:grant-type:jwt-bearer'
        })

        # Exchange the JWT for an access token.
        try:
            access_token_response = dtrequests.DTRequest.post(
                url='',
                base_url=token_url,
                data=request_data,
                headers={'Content-Type': 'application/x-www-form-urlencoded'},
                skip_auth=True,
            )
        except dterrors.BadRequest:
            # Re-raise exception with more specific information.
            raise dterrors.BadRequest(
                'Could not authenticate with the provided credentials.\n'
                'Read more: https://developer.d21s.com/docs/authentication'
                '/oauth2#common-errors'
            )

        # Return the access token in the request.
        return access_token_response



from disruptive.events.events import Event  # noqa
from disruptive.events.events import Touch  # noqa
from disruptive.events.events import Temperature  # noqa
from disruptive.events.events import ObjectPresent  # noqa
from disruptive.events.events import Humidity  # noqa
from disruptive.events.events import ObjectPresentCount  # noqa
from disruptive.events.events import TouchCount  # noqa
from disruptive.events.events import WaterPresent  # noqa
from disruptive.events.events import NetworkStatus  # noqa
from disruptive.events.events import NetworkStatusCloudConnector  # noqa
from disruptive.events.events import BatteryStatus  # noqa
from disruptive.events.events import LabelsChanged  # noqa
from disruptive.events.events import ConnectionStatus  # noqa
from disruptive.events.events import EthernetStatus  # noqa
from disruptive.events.events import CellularStatus  # noqa



from __future__ import annotations

# Project imports.
import disruptive.requests as dtrequests
from disruptive.outputs import OutputBase, Member


class Organization(OutputBase):
    """
    Represents an organization.

    When an organization response is received, the content
    is unpacked and the related attributes updated.

    Attributes
    ----------
    id : str
        Unique organization ID.
    display_name : str
        The provided display name.

    """

    def __init__(self, organization: dict) -> None:
        """
        Constructs the Organization object by unpacking the raw response.

        Parameters
        ----------
        organization : dict
            Unmodified organization response dictionary.

        """

        # Inherit from OutputBase parent.
        OutputBase.__init__(self, organization)

        # Unpack attributes from dictionary.
        self.id = organization['name'].split('/')[-1]
        self.display_name = organization['displayName']

    @classmethod
    def get_organization(cls,
                         organization_id: str,
                         **kwargs,
                         ) -> Organization:
        """
        Gets an organization specified by its ID.

        Parameters
        ----------
        organization_id : str
            Unique organization ID.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        organization : Organization
            Object representing the specified organization.

        """

        # Construct URL
        url = '/organizations/{}'.format(organization_id)

        # Return Organization object of GET request response.
        return cls(dtrequests.DTRequest.get(
            url=url,
            **kwargs,
        ))

    @classmethod
    def list_organizations(cls,
                           **kwargs,
                           ) -> list[Organization]:
        """
        List all available organizations.

        Parameters
        ----------
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        organizations : list[Organization]
            List of objects each representing an organization.

        """

        # Return list of Organization objects of paginated GET response.
        orgs = dtrequests.DTRequest.paginated_get(
            url='/organizations',
            pagination_key='organizations',
            **kwargs,
        )
        return [cls(org) for org in orgs]

    @staticmethod
    def list_members(organization_id: str,
                     **kwargs,
                     ) -> list[Member]:
        """
        List all members in the specified organization.

        Parameters
        ----------
        organization_id : str
            Unique ID of the organization to get members from.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        members : list[Member]
            List of objects each representing a member.

        """

        # Construct URL
        url = '/organizations/{}/members'.format(organization_id)

        # Return list of Member objects of paginated GET response.
        members = dtrequests.DTRequest.paginated_get(
            url=url,
            pagination_key='members',
            **kwargs,
        )
        return [Member(m) for m in members]

    @staticmethod
    def add_member(organization_id: str,
                   email: str,
                   roles: list[str],
                   **kwargs,
                   ) -> Member:
        """
        Add a new member to the specified organization.

        Parameters
        ----------
        organization_id : str
            Unique ID of the organization to add a member to.
        email : str
            Email of the user or Service Account to be added.
        roles : {["organization.admin"]} list[str]
            The role(s) to provide the new member in the organization.
            Currently only supports organization.admin.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        member : Member
            Object representing the newly added member.

        """

        # Construct URL
        url = '/organizations/{}/members'.format(organization_id)

        # Construct request body.
        body: dict = dict()
        body['roles'] = ['roles/' + r for r in roles]
        body['email'] = email

        # Return Member object of POST request response.
        return Member(dtrequests.DTRequest.post(
            url=url,
            body=body,
            **kwargs,
        ))

    @staticmethod
    def get_member(member_id: str,
                   organization_id: str,
                   **kwargs,
                   ) -> Member:
        """
        Get a member from the specified organization.

        Parameters
        ----------
        member_id : str
            Unique ID of the member to get.
            For Service Account members, this is the Service Account ID.
            For User members, this is the unique User ID.
        organization_id : str
            Unique ID of the organization to get a member from.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        member : Member
            Object representing the member.

        """

        # Construct URL
        url = '/organizations/{}/members/{}'.format(
            organization_id,
            member_id,
        )

        # Return Member object of GET request response.
        return Member(dtrequests.DTRequest.get(
            url=url,
            **kwargs,
        ))

    @staticmethod
    def remove_member(member_id: str,
                      organization_id: str,
                      **kwargs,
                      ) -> None:
        """
        Revoke a member's membership in the specified organization.

        Parameters
        ----------
        member_id : str
            Unique ID of the member to get.
            For Service Account members, this is the Service Account ID.
            For User members, this is the unique User ID.
        organization_id : str
            Unique ID of the organization to remove a member from.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        """

        # Construct URL
        url = '/organizations/{}/members/{}'.format(
            organization_id,
            member_id,
        )

        # Send DELETE request, but return nothing.
        dtrequests.DTRequest.delete(
            url=url,
            **kwargs,
        )

    @staticmethod
    def get_member_invite_url(member_id: str,
                              organization_id: str,
                              **kwargs,
                              ) -> None:
        """
        Get the invite URL for a member with pending invite.

        This will only work if the invite is still pending. If the invite has
        already been accepted, an error is raised.

        Parameters
        ----------
        member_id : str
            Unique ID of the member to get.
            For Service Account members, this is the Service Account ID.
            For User members, this is the unique User ID.
        organization_id : str
            Unique ID of the organization of the member to get the URL from.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Raises
        ------
        BadRequest
            If the invite has already been accepted.

        """

        # Construct URL
        url = '/organizations/{}/members/{}'.format(
            organization_id,
            member_id,
        ) + ':getInviteUrl'

        # Return url string in GET response.
        return dtrequests.DTRequest.get(
            url=url,
            **kwargs,
        )['inviteUrl']

    @staticmethod
    def list_permissions(organization_id: str,
                         **kwargs,
                         ) -> list[str]:
        """
        List permissions available in the specified organization.

        Parameters
        ----------
        organization_id : str
            Unique ID of the organization to list permissions in.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        permissions : list[str]
            List of available permissions.

        """

        # Construct URL
        url = '/organizations/{}/permissions'.format(organization_id)

        # Return list of permissions in GET response.
        return dtrequests.DTRequest.paginated_get(
            url=url,
            pagination_key='permissions',
            **kwargs,
        )



from __future__ import annotations

# Standard library imports.
from typing import Optional

# Project imports.
import disruptive.requests as dtrequests
import disruptive.outputs as dtoutputs
import disruptive.transforms as dttrans


class ServiceAccount(dtoutputs.OutputBase):
    """
    Represents a serviceaccount.

    When a serviceaccount response is received, the content is
    unpacked and the related attributes are updated.

    Attributes
    ----------
    serviceaccount_id : str
        Unique serviceaccount ID.
    email : str
        Unique serviceaccount email.
    display_name : str
        The provided display name.
    basic_auth : bool
        True if Basic Auth is enabled, otherwise False.
    create_time : datetime
        Timestamp of when the serviceaccount was created.
    update_time : datetime
        Timestamp of when the serviceaccount was last updated.

    """

    def __init__(self, serviceaccount: dict) -> None:
        """
        Constructs the ServiceAccount object by unpacking the raw response.

        Parameters
        ----------
        serviceaccount : dict
            Unmodified serviceaccount response dictionary.

        """

        # Inherit from Response parent.
        dtoutputs.OutputBase.__init__(self, serviceaccount)

        # Unpack attributes from dictionary.
        self.serviceaccount_id = serviceaccount['name'].split('/')[-1]
        self.email = serviceaccount['email']
        self.display_name = serviceaccount['displayName']
        self.basic_auth = serviceaccount['enableBasicAuth']
        self.create_time = dttrans.to_datetime(serviceaccount['createTime'])
        self.update_time = dttrans.to_datetime(serviceaccount['updateTime'])

    @classmethod
    def get_serviceaccount(cls,
                           serviceaccount_id: str,
                           project_id: str,
                           **kwargs,
                           ) -> ServiceAccount:
        """
        Gets a serviceaccount specified by its ID.

        Parameters
        ----------
        serviceaccount_id : str
            Unique ID of the target serviceaccount.
        project_id : str
            Unique ID of the target project.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        serviceaccount : ServiceAccount
            Object representing the target serviceaccount.

        """

        # Construct URL.
        url = '/projects/{}/serviceaccounts/{}'.format(
            project_id,
            serviceaccount_id,
        )

        # Return ServiceAccount object of GET request response.
        return cls(dtrequests.DTRequest.get(
            url=url,
            **kwargs,
        ))

    @classmethod
    def list_serviceaccounts(cls,
                             project_id: str,
                             **kwargs,
                             ) -> list[ServiceAccount]:
        """
        List all available serviceaccounts in the specified project.

        Parameters
        ----------
        project_id : str
            Unique ID of the target project.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        serviceaccounts : list[ServiceAccount]
            List of objects each representing a serviceaccount.

        """

        # Construct URL.
        url = '/projects/{}/serviceaccounts'.format(project_id)

        # Return list of ServiceAccount objects of paginated GET response.
        service_accounts = dtrequests.DTRequest.paginated_get(
            url=url,
            pagination_key='serviceAccounts',
            **kwargs,
        )
        return [cls(sa) for sa in service_accounts]

    @classmethod
    def create_serviceaccount(cls,
                              project_id: str,
                              display_name: str = '',
                              basic_auth: bool = False,
                              **kwargs,
                              ) -> ServiceAccount:
        """
        Create a new serviceaccount in the specified project.

        Parameters
        ----------
        project_id : str
            Unique ID of the target project.
        display_name : str, optional
            Sets a display name for the serviceaccount.
        basic_auth : bool, optional
            Enables Basic Auth for the serviceaccount if True.
            Defaults to False.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        serviceaccount : ServiceAccount
            Object representing the newly created serviceaccount.

        """

        # Construct URL.
        url = '/projects/{}/serviceaccounts'.format(project_id)

        # Construct body.
        body: dict = dict()
        body['enableBasicAuth'] = basic_auth
        if len(display_name) > 0:
            body['displayName'] = display_name

        # Return ServiceAccount object of GET request response.
        return cls(dtrequests.DTRequest.post(
            url=url,
            body=body,
            **kwargs,
        ))

    @classmethod
    def update_serviceaccount(cls,
                              serviceaccount_id: str,
                              project_id: str,
                              display_name: Optional[str] = None,
                              basic_auth: Optional[bool] = None,
                              **kwargs,
                              ) -> ServiceAccount:
        """
        Updates the attributes of a specified serviceaccount.

        Parameters
        ----------
        serviceaccount_id : str
            Unique ID of the target serviceaccount.
        project_id : str
            Unique ID of the target project.
        display_name : str, optional
            Updates the serviceaccount display name.
        basic_auth : bool, optional
            If True, enables Basic Auth while False disables it.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        """

        # Construct URL.
        url = '/projects/{}/serviceaccounts/{}'.format(
            project_id,
            serviceaccount_id,
        )

        # Construct body.
        body: dict = dict()
        if display_name is not None:
            body['displayName'] = display_name
        if basic_auth is not None:
            body['enableBasicAuth'] = basic_auth

        # Return ServiceAccount object of GET request response.
        return cls(dtrequests.DTRequest.patch(
            url=url,
            body=body,
            **kwargs,
        ))

    @classmethod
    def delete_serviceaccount(cls,
                              serviceaccount_id: str,
                              project_id: str,
                              **kwargs,
                              ) -> None:
        """
        Deletes the specified serviceaccount.

        Parameters
        ----------
        serviceaccount_id : str
            Unique ID of the serviceaccount to delete.
        project_id : str
            Unique ID of the target project.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        """

        # Construct URL.
        url = '/projects/{}/serviceaccounts/{}'.format(
            project_id,
            serviceaccount_id,
        )

        # Send DELETE request, but return nothing.
        dtrequests.DTRequest.delete(
            url=url,
            **kwargs,
        )

    @staticmethod
    def get_key(serviceaccount_id: str,
                project_id: str,
                key_id: str,
                **kwargs,
                ) -> Key:
        """
        Get the key of a serviceaccount.

        Parameters
        ----------
        serviceaccount_id : str
            Unique ID of the target serviceaccount.
        project_id : str
            Unique ID of the target project.
        key_id : str
            Unique ID of the target key.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        key : Key
            Object representing the target key.

        """

        # Construct URL.
        url = '/projects/{}/serviceaccounts/{}/keys/{}'.format(
            project_id,
            serviceaccount_id,
            key_id,
        )

        # Return Key object of GET request response.
        return Key(dtrequests.DTRequest.get(
            url=url,
            **kwargs,
        ))

    @staticmethod
    def list_keys(serviceaccount_id: str,
                  project_id: str,
                  **kwargs,
                  ) -> list[Key]:
        """
        Get a list of all keys for a serviceaccount.

        Parameters
        ----------
        serviceaccount_id : str
            Unique ID of the target serviceaccount.
        project_id : str
            Unique ID of the target project.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        keys : list[Key]
            List of objects each representing a key.

        """

        # Construct URL.
        url = '/projects/{}/serviceaccounts/{}/keys'.format(
            project_id,
            serviceaccount_id,
        )

        # Return list of Key objects of paginated GET response.
        keys = dtrequests.DTRequest.paginated_get(
            url=url,
            pagination_key='keys',
            **kwargs,
        )
        return [Key(key) for key in keys]

    @staticmethod
    def create_key(serviceaccount_id: str,
                   project_id: str,
                   **kwargs,
                   ) -> Key:
        """
        Create a new key for the specified serviceaccount.

        Parameters
        ----------
        serviceaccount_id : str
            Unique ID of the target serviceaccount.
        project_id : str
            Unique ID of the target project.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        key : Key
            Object representing the newly created key.

        """

        # Construct URL.
        url = '/projects/{}/serviceaccounts/{}/keys'.format(
            project_id,
            serviceaccount_id,
        )

        # Return Key object of POST request response.
        response = dtrequests.DTRequest.post(
            url=url,
            **kwargs,
        )
        return Key._with_secret(response)

    @staticmethod
    def delete_key(serviceaccount_id: str,
                   key_id: str,
                   project_id: str,
                   **kwargs,
                   ) -> None:
        """
        Deletes a key in the specified serviceaccount.

        Parameters
        ----------
        serviceaccount_id : str
            Unique ID of the target serviceaccount.
        key_id : str
            Unique ID of the key to delete.
        project_id : str
            Unique ID of the target project.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        """

        # Construct URL.
        url = '/projects/{}/serviceaccounts/{}/keys/{}'.format(
            project_id,
            serviceaccount_id,
            key_id,
        )

        # Send DELETE request, but return nothing.
        dtrequests.DTRequest.delete(
            url=url,
            **kwargs,
        )


class Key(dtoutputs.OutputBase):
    """
    Represents a key in a serviceaccount.

    When a key response is received, the content is
    unpacked and the related attributes are updated.

    Attributes
    ----------
    key_id : str
        Unique key ID.
    secret : str, None
        If the Key object was constructed from a newly created key, i.e. from
        calling the create_key() method, this attributes contains the
        key secret. This only displays once, and is otherwise None.
    create_time : datetime
        Timestamp of when the key was created.

    """

    def __init__(self, key: dict) -> None:
        """
        Constructs the Key object by unpacking the raw key response.

        Parameters
        ----------
        key : dict
            Key response dictionary.

        """
        # Inherit from Response parent.
        dtoutputs.OutputBase.__init__(self, key)

        # Initialize secret, which is only not-None when created.
        self.secret = None

        # Unpack attributes from dictionary.
        self.key_id = key['id']
        self.create_time = dttrans.to_datetime(key['createTime'])
        if 'secret' in key:
            self.secret = key['secret']

    @classmethod
    def _with_secret(cls, key: dict) -> Key:
        """
        Moves the secret field in the response dictionary.

        This is done for convenience and consistency. For some reason, when
        the secret is included, the raw response dictionary is changed, adding
        another layer outside the key field. This just flattens the structure.

        Parameters
        ----------
        key : dict
            Unmodified key response dictionary.

        Returns
        -------
        flattened : dict
            Modified key response dictionary.

        """

        flattened = key['key']
        flattened['secret'] = key['secret']
        return cls(flattened)



from __future__ import annotations

# Standard library imports.
from typing import Optional

# Project imports.
import disruptive
import disruptive.requests as dtrequests
import disruptive.outputs as dtoutputs
import disruptive.dataconnector_configs.dataconnector_configs as dcon_configs
from disruptive.outputs import Metric


class DataConnector(dtoutputs.OutputBase):
    """
    Represents a dataconnector.

    When a dataconnector response is received, the content
    is unpacked and the related attributes are updated.

    Attributes
    ----------
    dataconnector_id : str
        Unique dataconnector ID.
    project_id : str
        Unique ID of the project where the dataconnector recides.
    display_name : str
        The provided display name.
    dataconnector_type : str
        Dataconnector type. Currently only HTTP_PUSH is available.
    status : str
        Whether the dataconnector is
        "ACTIVE", "USER_DISABLED", or "SYSTEM_DISABLED".
    config : HttpPush, None
        An object representing the type-specific configuration.
    event_Types : list[str]
        List of event types that should be forwarded.
        If empty, all event types are forwarded.
    labels : list[str]
        Device labels are not included by default and will only be forwarded
        with the event if the key is specified in this list.

    """

    def __init__(self, dataconnector: dict) -> None:
        """
        Constructs the DataConnector object by unpacking
        the raw dataconnector response.

        Parameters
        ----------
        dataconnector : dict
            Unmodified dataconnector response dictionary.

        """

        # Inherit from OutputBase parent.
        dtoutputs.OutputBase.__init__(self, dataconnector)

        # Unpack attributes from dictionary.
        self.dataconnector_id = dataconnector['name'].split('/')[-1]
        self.project_id = dataconnector['name'].split('/')[1]
        self.status = dataconnector['status']
        self.display_name = dataconnector['displayName']
        self.event_types = dataconnector['events']
        self.labels = dataconnector['labels']
        self.dataconnector_type = dataconnector['type']
        self.config = dcon_configs._from_dict(dataconnector)

    @classmethod
    def get_dataconnector(cls,
                          dataconnector_id: str,
                          project_id: str,
                          **kwargs,
                          ) -> DataConnector:
        """
        Gets a dataconnector specified by its ID.

        Parameters
        ----------
        dataconnector_id : str
            Unique dataconnector ID.
        project_id : str
            Unique project ID.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        dataconnector : DataConnector
            Object representing the specified dataconnector.

        """

        # Construct URL
        url = '/projects/{}/dataconnectors/{}'
        url = url.format(project_id, dataconnector_id)

        # Return DataConnector object of GET request response.
        return cls(dtrequests.DTRequest.get(
            url=url,
            **kwargs,
        ))

    @classmethod
    def list_dataconnectors(cls,
                            project_id: str,
                            **kwargs,
                            ) -> list[DataConnector]:
        """
        List all available dataconnectors in the specified project.

        Parameters
        ----------
        project_id : str
            Unique project ID.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        dataconnectors : list[DataConnector]
            List of objects each representing a dataconnector.

        """

        # Return list of DataConnector objects of paginated GET response.
        dataconnectors = dtrequests.DTRequest.paginated_get(
            url='/projects/{}/dataconnectors'.format(project_id),
            pagination_key='dataConnectors',
            **kwargs,
        )
        return [cls(dcon) for dcon in dataconnectors]

    @classmethod
    def create_dataconnector(cls,
                             project_id: str,
                             config: disruptive.dataconnector_configs.HttpPush,
                             display_name: str = '',
                             status: str = 'ACTIVE',
                             events: list[str] = [],
                             labels: list[str] = [],
                             **kwargs,
                             ) -> DataConnector:
        """
        Create a new dataconnector in the specified project.

        Parameters
        ----------
        project_id : str
            Unique project ID.
        config : HttpPush
            An object representing the type-specific configuration.
        display_name : str, optional
            Sets a display name for the project.
        status : {"ACTIVE", "USER_DISABLED"} strm optional
            Status of the new dataconnector.
        events : list[str], optional
            List of event types the dataconnectors should forward.
        labels : list[str], optional
            List of labels to forward with each event.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        dataconnector : DataConnector
            Object representing the newly created dataconnector.

        """

        # Construct request body dictionary.
        body: dict = dict()
        body['status'] = status
        body['events'] = events
        body['labels'] = labels
        if len(display_name) > 0:
            body['displayName'] = display_name

        # Add the appropriate field depending on config.
        body['type'] = config.dataconnector_type
        key, value = config._to_dict()
        body[key] = value

        # Construct URL.
        url = '/projects/{}/dataconnectors'.format(project_id)

        # Return DataConnector object of POST request response.
        return cls(dtrequests.DTRequest.post(
            url=url,
            body=body,
            **kwargs,
        ))

    @classmethod
    def update_dataconnector(
        cls,
        dataconnector_id: str,
        project_id: str,
        config: Optional[disruptive.dataconnector_configs.HttpPush] = None,
        display_name: Optional[str] = None,
        status: Optional[str] = None,
        events: Optional[list[str]] = None,
        labels: Optional[list[str]] = None,
        **kwargs,
    ) -> DataConnector:
        """
        Updates the attributes of a dataconnector.

        Parameters
        ----------
        dataconnector_id : str
            Unique ID of the dataconnector to update.
        project_id : str
            Unique ID of the project that contains the dataconnector.
        config : HttpPush, None
            An object representing the type-specific configuration.
        display_name : str, optional
            Sets a display name for the dataconnector.
        status : {"ACTIVE", "USER_DISABLED"} str, optional
            Status of the dataconnector.
        events : list[str], optional
            List of event types the dataconnectors should forward.
        labels : list[str], optional
            List of labels to forward with each event.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        """

        # Construct request body dictionary.
        body: dict = dict()
        if display_name is not None:
            body['displayName'] = display_name
        if status is not None:
            body['status'] = status
        if events is not None:
            body['events'] = events
        if labels is not None:
            body['labels'] = labels

        # Add the appropriate field depending on config.
        if config is not None:
            key, value = config._to_dict()
            body[key] = value

        # Construct URL.
        url = '/projects/{}/dataconnectors/{}'
        url = url.format(project_id, dataconnector_id)

        # Return DataConnector object of PATCH request response.
        return cls(dtrequests.DTRequest.patch(
            url=url,
            body=body,
            **kwargs,
        ))

    @classmethod
    def delete_dataconnector(cls,
                             dataconnector_id: str,
                             project_id: str,
                             **kwargs,
                             ) -> None:
        """
        Deletes the specified dataconnector.

        Parameters
        ----------
        dataconnector_id : str
            Unique ID of the dataconnector to delete.
        project_id : str
            Unique ID of the project that contains the dataconnector.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        """

        # Construct URL.
        url = '/projects/{}/dataconnectors/{}'
        url = url.format(project_id, dataconnector_id)

        # Send DELETE request, but return nothing.
        dtrequests.DTRequest.delete(
            url=url,
            **kwargs,
        )

    @classmethod
    def get_metrics(cls,
                    dataconnector_id: str,
                    project_id: str,
                    **kwargs,
                    ) -> Metric:
        """
        Get the metrics of the last 3 hours for a dataconnector.

        Parameters
        ----------
        dataconnector_id : str
            Unique ID of the dataconnector to list metrics for.
        project_id : str
            Unique ID of the project that contains the dataconnector.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        metric : Metric
            Object representing the fetched metrics.

        """

        # Construct URL.
        url = '/projects/{}/dataconnectors/{}'
        url = url.format(project_id, dataconnector_id)
        url += ':metrics'

        # Return Metric object of GET request response.
        return Metric(dtrequests.DTRequest.get(
            url=url,
            **kwargs,
        ))

    @classmethod
    def sync_dataconnector(cls,
                           dataconnector_id: str,
                           project_id: str,
                           **kwargs,
                           ) -> None:
        """
        Synchronizes the current dataconnector state.

        This method let's you synchronize your cloud service with the current
        state of the devices in your project. This entails pushing the most
        recent event of each type for all devices in your project.

        Parameters
        ----------
        dataconnector_id : str
            Unique ID of the dataconnector to synchronize.
        project_id : str
            Unique ID of the project that contains the dataconnector.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        """

        # Construct URL.
        url = '/projects/{}/dataconnectors/{}'
        url = url.format(project_id, dataconnector_id)
        url += ':sync'

        # Send POST request, but return nothing.
        dtrequests.DTRequest.post(
            url=url,
            **kwargs,
        )



touch = 'touch'
temperature = 'temperature'
object_present = 'objectPresent'
humidity = 'humidity'
object_present_count = 'objectPresentCount'
touch_count = 'touchCount'
water_present = 'waterPresent'
battery_status = 'batteryStatus'
network_status = 'networkStatus'
labels_changed = 'labelsChanged'
connection_status = 'connectionStatus'
ethernet_status = 'ethernetStatus'
cellularStatus = 'cellularStatus'



from __future__ import annotations

# Standard library imports
from typing import Generator, Optional

# Project Imports.
import disruptive.requests as dtrequests
from disruptive.events.events import Event
from disruptive.authentication import Auth


class Stream():
    """
    Contains staticmethods for streaming events.
    Used for namespacing only and thus does not have a constructor

    """

    @staticmethod
    def device(device_id: str,
               project_id: str,
               event_types: Optional[list[str]] = None,
               **kwargs,
               ) -> Generator:
        """
        Streams events for a single device.

        Implements a basic retry-routine. If connection is lost, the stream
        will attempt to reconnect with an exponential backoff. Potential
        lost events while reconnecting are, however, not acocunted for.

        Parameters
        ----------
        device_id : str
            Unique ID of the target device.
        project_id : str
            Unique ID of the target project.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.

        Returns
        -------
        stream : Generator
            A python Generator type that yields each new event in the stream.

        Examples
        --------
        >>> for event in dt.Stream.device(project_id, device_id):
        ...     print(event)

        """

        # Construct URL.
        url = '/projects/{}/devices/{}:stream'.format(
            project_id,
            device_id
        )

        # Construct parameters dictionary.
        params: dict = dict()
        if event_types is not None:
            params['event_types'] = event_types

        # Relay generator output.
        for event in dtrequests.DTRequest.stream(url, params=params, **kwargs):
            yield Event(event)

    @staticmethod
    def project(project_id: str,
                device_ids: Optional[list[str]] = None,
                label_filters: Optional[list[str]] = None,
                device_types: Optional[list[str]] = None,
                event_types: Optional[list[str]] = None,
                auth: Optional[Auth] = None,
                ) -> Generator:
        """
        Streams events for a multiple devices in a project.

        Implements a basic retry-routine. If connection is lost, the stream
        will attempt to reconnect with an exponential backoff. Potential
        lost events while reconnecting are, however, not acocunted for.

        Parameters
        ----------
        project_id : str
            Unique ID of the target project.
        device_ids : list[str], optional
            Only includes events from the specified device(s).
        label_filters : list[str], optional
            Only includes events from devices with specified label(s).
        device_types : list[str], optional
            Only includes events from devices with specified type(s).
        event_types : list[str], optional
            Only includes events of the specified type(s).
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.

        Returns
        -------
        stream : Generator
            A python Generator type that yields each new event in the stream.

        Examples
        --------
        >>> for event in dt.Stream.project(project_id):
        ...     print(event)

        """

        # Construct parameters dictionary.
        params: dict = dict()
        if device_ids is not None:
            params['device_ids'] = device_ids
        if device_types is not None:
            params['device_types'] = device_types
        if label_filters is not None:
            params['label_filters'] = label_filters
        if event_types is not None:
            params['event_types'] = event_types

        # Relay generator output.
        url = '/projects/{}/devices:stream'.format(project_id)
        for event in dtrequests.DTRequest.stream(url, params=params):
            yield Event(event)



from __future__ import annotations

# Project imports.
import disruptive.requests as dtrequests
from disruptive.outputs import OutputBase


class Role(OutputBase):
    """
    Represents a role.

    When a role response is received, the content is
    unpacked and the related attributes are updated.

    Attributes
    ----------
    role : str
        Name of the role.
    display_name : str
        Display name of the role.
    description : str
        Description of what the role entails.
    permissions : list[str]
        List of permissions provided with the role.

    """

    def __init__(self, role: dict) -> None:
        """
        Constructs the Role object by unpacking the raw role response.

        Parameters
        ----------
        role : dict
            Unmodified role response dictionary.

        """

        # Inherit from Response parent.
        OutputBase.__init__(self, role)

        # Unpack attributes from dictionary.
        self.role = role['name'].split('/')[-1]
        self.display_name = role['displayName']
        self.description = role['description']
        self.permissions = role['permissions']

    @classmethod
    def get_role(cls,
                 role: str,
                 **kwargs,
                 ) -> Role:
        """
        Gets a role specified by its name.

        Parameters
        ----------
        role : str
            Name of the role.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        role : Role
            Object representing the specified role.

        """

        # Return list of Role objects.
        return cls(dtrequests.DTRequest.get(
            url='/roles/' + role,
            **kwargs,
        ))

    @classmethod
    def list_roles(cls,
                   **kwargs,
                   ) -> list[Role]:
        """
        List all available roles.

        Parameters
        ----------
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        roles : list[Role]
            List of objects each representing a role.

        """

        # Return list of Role objects.
        response = dtrequests.DTRequest.paginated_get(
            url='/roles',
            pagination_key='roles',
            **kwargs,
        )
        return [cls(r) for r in response]



# Project imports
import disruptive.log as dtlog

# Standard library imports.
from typing import Optional


def _from_dict(dataconnector: dict):
    # Isolate the dataconnector type.
    dataconnector_type = dataconnector['type']

    # Select the appropriate config depending on type.
    if dataconnector_type == 'HTTP_PUSH':
        # Create and return an HttpPush object.
        return HttpPush(
            url=dataconnector['httpConfig']['url'],
            signature_secret=dataconnector['httpConfig']['signatureSecret'],
            headers=dataconnector['httpConfig']['headers'],
        )
    else:
        # If this else statement runs, no config is (yet) available for type.
        dtlog.log('No config available for {} dataconnectors.'.format(
            dataconnector_type
        ))


class HttpPush():
    """
    Type-specific configurations for the HTTP_PUSH dataconnector.

    Attributes
    ----------
    url : str
        Endpoint URL towards which events are forwarded. Must be HTTPS.
    signature_secret : str
        Secret with which each forwarded event is signed.
    headers : dict[str, str]
        Dictionary of headers to include with each forwarded event.

    """

    dataconnector_type = 'HTTP_PUSH'

    def __init__(self,
                 url: Optional[str] = None,
                 signature_secret: Optional[str] = None,
                 headers: Optional[dict] = None,
                 ) -> None:
        """
        Constructs the HttpPush object.

        Parameters
        ----------
        url : str, optional
            Endpoint URL towards which events are forwarded. Must be HTTPS.
        signature_secret : str, optional
            Secret with which each forwarded event is signed.
        headers : dict[str, str], optional
            Dictionary of headers to include with each forwarded event.

        """

        # Set parameter attributes.
        self.url = url
        self.signature_secret = signature_secret
        self.headers = headers

    def _to_dict(self):
        config: dict = dict()
        if self.url is not None:
            config['url'] = self.url
        if self.signature_secret is not None:
            config['signatureSecret'] = self.signature_secret
        if self.headers is not None:
            config['headers'] = self.headers
        return 'httpConfig', config



from __future__ import annotations

# Standard library imports.
import time
import json
from typing import Optional

# Third-party imports.
import requests

# Project imports
import disruptive as dt
import disruptive.log as dtlog
import disruptive.errors as dterrors


class DTRequest():

    def __init__(self, method, url, **kwargs):
        # Set attributes from parameters.
        self.method = method
        self.url = url

        # Set default attributes, though many are updated in _unpack_kwargs().
        self.base_url = dt.api_url
        self.params = {}
        self.headers = {}
        self.body = None
        self.data = None
        self.request_timeout = dt.request_timeout
        self.request_retries = dt.request_retries
        self.log = dt.log

        # Unpack kwargs and set attributes thereafter.
        self._unpack_kwargs(**kwargs)

        # Check that provided arguments are valid.
        self._sanitize_arguments()

        # Construct full url from base and target endpoint.
        self.full_url = self.base_url + self.url

    def _unpack_kwargs(self, **kwargs):
        if 'params' in kwargs:
            self.params = kwargs['params']
        if 'headers' in kwargs:
            self.headers = kwargs['headers']
        if 'body' in kwargs:
            self.body = kwargs['body']
        if 'data' in kwargs:
            self.data = kwargs['data']

        # Check if request_timeout is overriden.
        if 'request_timeout' in kwargs:
            self.request_timeout = kwargs['request_timeout']

        # Check if request_retries is overriden.
        if 'request_retries' in kwargs:
            self.request_retries = kwargs['request_retries']

        # Check if base_url is overriden.
        if 'base_url' in kwargs:
            self.base_url = kwargs['base_url']

        # Check if log is overriden.
        if 'log' in kwargs:
            self.log = kwargs['log']

        # Add authorization header to request except when explicitly otherwise.
        if 'skip_auth' not in kwargs or kwargs['skip_auth'] is False:
            # If provided, override the package-wide auth with provided object.
            if 'auth' in kwargs:
                self.headers['Authorization'] = kwargs['auth'].get_token()
            # If not, use package-wide auth object.
            else:
                self.headers['Authorization'] = dt.default_auth.get_token()

    def _sanitize_arguments(self):
        # Check that request_timeout > 0.
        if self.request_timeout <= 0:
            raise dterrors.ConfigurationError(
                'Configuration parameter request_timeout has value {}, but '
                'must be float greater than 0.'.format(self.request_timeout)
            )

        # Check that request_retries > 0.
        if self.request_retries <= 0:
            raise dterrors.ConfigurationError(
                'Configuration parameter request_retries has value {}, but '
                'must be integer greater than 0.'.format(self.request_retries)
            )

    def _request_wrapper(self,
                         method: str,
                         url: str,
                         params: dict,
                         headers: dict,
                         body: Optional[dict],
                         data: Optional[str],
                         timeout: int,
                         ):
        # Attempt to send the request.
        try:
            # Use the requests package to send the request.
            res = requests.request(
                method=method,
                url=url,
                params=params,
                headers=headers,
                json=body,
                data=data,
                timeout=timeout
            )

            # Isolate the data of interest in the response.
            return DTResponse(res.json(), res.status_code, res.headers), None

        except requests.exceptions.RequestException as e:
            return DTResponse({}, None, {}), e
        except ValueError as e:
            # Requests' .json() fails when no json is returned (code 405).
            return DTResponse({}, res.status_code, res.headers), e

    def _send_request(self, nth_attempt=1):
        # Log the request.
        dtlog.log('Request [{}] to {}.'.format(
            self.method,
            self.base_url + self.url
        ), override=self.log)

        res, req_error = self._request_wrapper(
            method=self.method,
            url=self.full_url,
            params=self.params,
            headers=self.headers,
            body=self.body,
            data=self.data,
            timeout=self.request_timeout,
        )

        # Log the response.
        dtlog.log('Response [{}].'.format(
            res.status_code
        ), override=self.log)

        # If _request_wrapper raised an exception, the request failed.
        if req_error is not None:
            error, should_retry, retry_after = self._parse_request_error(
                req_error, res.data, nth_attempt
            )
        else:
            # Parse the status_code and select an appropriate error.
            # If there is any hope at all that a retry might resolve the error,
            # should_retry will be True. (eg. a 401).
            error, should_retry, retry_after = self._parse_api_status_code(
                res.status_code, res.data, res.headers, nth_attempt
            )

        # Check if retry is required.
        if should_retry and nth_attempt < self.request_retries:

            dtlog.log("Got error {}. Will retry up to {} more times".format(
                error,
                dt.request_retries - nth_attempt,
            ), override=self.log)

            # Sleep if necessary.
            if retry_after is not None:
                time.sleep(retry_after)

            # Attempt the request again recursively, iterating counter.
            res.data = self._send_request(nth_attempt+1)

        else:
            # If set, raise the error chosen by dterrors.parse_error().
            if error is not None:
                raise error

        return res.data

    def _parse_request_error(self, caught_error, data, nth_attempt):
        # Read Timeouts should be attempted again.
        if isinstance(caught_error, requests.exceptions.ReadTimeout):
            return dterrors.ReadTimeout, True, nth_attempt**2

        # Connection errors should be attempted again.
        elif isinstance(caught_error, requests.exceptions.ConnectionError):
            return (
                ConnectionError('Failed to establish connection.'),
                True,
                nth_attempt**2,
            )
        else:
            # Uncategorized error has been raised.
            return dterrors.UnknownError(data), False, None

    def _parse_api_status_code(self, status_code, data, headers, nth_attempt):
        # Check for API errors.
        if status_code == 200:
            return None, False, None
        elif status_code == 400:
            return dterrors.BadRequest(data), False, None
        elif status_code == 401:
            # The first retry is #1. Therefor, retry_count < 2 will
            # result in a a single retry attempt.
            return dterrors.Unauthorized(data), nth_attempt < 2, None
        elif status_code == 403:
            return dterrors.Forbidden(data), False, None
        elif status_code == 404:
            return dterrors.NotFound(data), False, None
        elif status_code == 409:
            return dterrors.Conflict(data), False, None
        elif status_code == 429:
            if 'Retry-After' in headers:
                return (
                    dterrors.TooManyRequests(data),
                    True,
                    int(headers['Retry-After']),
                )
            else:
                return dterrors.TooManyRequests(data), False, None
        elif status_code == 500:
            return dterrors.InternalServerError(data), True, nth_attempt**2
        elif status_code == 503:
            return dterrors.InternalServerError(data), True, nth_attempt**2
        elif status_code == 504:
            return dterrors.InternalServerError(data), True, nth_attempt**2 + 9
        else:
            return dterrors.UnknownError(data), False, None

    @classmethod
    def get(cls, url, **kwargs):
        req = cls('GET', url, **kwargs)
        return req._send_request()

    @classmethod
    def post(cls, url, **kwargs):
        req = cls('POST', url, **kwargs)
        return req._send_request()

    @classmethod
    def patch(cls, url, **kwargs):
        req = cls('PATCH', url, **kwargs)
        return req._send_request()

    @classmethod
    def delete(cls, url, **kwargs):
        req = cls('DELETE', url, **kwargs)
        return req._send_request()

    @classmethod
    def paginated_get(cls,
                      url: str,
                      pagination_key: str,
                      params: dict[str, str] = {},
                      **kwargs
                      ):
        # Initialize output list.
        results = []

        # Loop until paging has finished.
        while True:
            response = cls.get(url, params=params, **kwargs)
            results += response[pagination_key]

            if len(response['nextPageToken']) > 0:
                params['pageToken'] = response['nextPageToken']
            else:
                break

        return results

    @classmethod
    def stream(cls, url: str, **kwargs):
        # Set ping constants.
        PING_INTERVAL = 10
        PING_JITTER = 2

        # Expand url with base_url.
        url = dt.api_url + url

        # Set error variable that if not None, raise it.
        error = None

        def stream_retry_logic(nth_attempt):
            # Print the error and try again up to max_request_retries.
            if nth_attempt < request_retries:
                dtlog.log('Connection lost. Retry {}/{}.'.format(
                    nth_attempt+1,
                    request_retries,
                ))

                # Exponential backoff in sleep time.
                time.sleep(2**nth_attempt)

                return nth_attempt + 1, None
            else:
                return nth_attempt, dterrors.ConnectionError(
                    'Stream retry attempts has been exhausted.'
                )

        # Unpack kwargs.
        params = kwargs['params'] if 'params' in kwargs else {}
        headers = kwargs['headers'] if 'headers' in kwargs else {}
        if 'request_retries' in kwargs:
            request_retries = kwargs['request_retries']
        else:
            request_retries = dt.request_retries

        # Add ping parameter to dictionary.
        params['ping_interval'] = str(PING_INTERVAL) + 's'

        # If provided, override package-wide auth with argument.
        if 'auth' in kwargs:
            headers['Authorization'] = kwargs['auth'].get_token()
        else:
            headers['Authorization'] = dt.default_auth.get_token()

        # Set up a simple catch-all retry policy.
        nth_attempt = 0
        while nth_attempt <= request_retries:
            # Check if error is set.
            if error is not None:
                raise error

            try:
                # Set up a stream connection.
                # Connection will timeout and reconnect if no single event
                # is received in an interval of ping_interval + ping_jitter.
                dtlog.log('Starting stream...')
                stream = requests.get(
                    url=url,
                    stream=True,
                    timeout=PING_INTERVAL + PING_JITTER,
                    params=params,
                    headers=headers,
                )

                # Iterate through the events as they come in (one per line).
                for line in stream.iter_lines():
                    # Decode the response payload and break on error.
                    payload = json.loads(line.decode('ascii'))
                    if 'result' not in payload:
                        break

                    # Reset retry counter.
                    nth_attempt = 0

                    # Check for ping event.
                    event = payload['result']['event']
                    if event['eventType'] == 'ping':
                        dtlog.log('Ping received.')
                        continue

                    # Yield event to generator.
                    yield event

            except KeyboardInterrupt:
                break

            except requests.exceptions.ReadTimeout:
                nth_attempt, error = stream_retry_logic(nth_attempt)

            except requests.exceptions.ConnectionError:
                nth_attempt, error = stream_retry_logic(nth_attempt)


class DTResponse():

    def __init__(self, data, status_code, headers):
        self.data = data
        self.status_code = status_code
        self.headers = headers



from __future__ import annotations

# Standard library imports.
from typing import Optional
from datetime import datetime

# Project imports.
import disruptive.requests as dtrequests
import disruptive.outputs as dtoutputs
import disruptive.transforms as dttrans
from disruptive.events.events import Event


class EventHistory(dtoutputs.OutputBase):
    """
    Contains staticmethods for fetching event history.
    Used for namespacing only and thus does not have a constructor

    Attributes
    ----------
    events_list : list[Event]
        List of all events fetched in history.

    """

    def __init__(self, events_list: list[Event]) -> None:
        """
        Constructs the EventHistory object from a list of events.

        Parameters
        ----------
        events_list : list[Event]
            A list of event objects, each representing a single event.

        """

        # Inherit from OutputBase parent.
        dtoutputs.OutputBase.__init__(self, {})

        # Set parameter attributes.
        self.events_list = events_list

    @classmethod
    def list_events(cls,
                    device_id: str,
                    project_id: str,
                    event_types: Optional[list[str]] = None,
                    start_time: Optional[str | datetime] = None,
                    end_time: Optional[str | datetime] = None,
                    **kwargs,
                    ) -> EventHistory:
        """
        Get the event history for a single device.

        Parameters
        ----------
        device_id : str
            Unique ID of the target device.
        project_id : str
            Unique ID of the target project.
        event_types : list[str], optional
            If provided, only the specified event types are fetched.
        start_time : str, datetime, optional
            Specifies from when event history is fetched.
            Defaults to 24 hours ago.
        end_time : str, datetime, optional
            Specified until when event history is fetched.
            Defaults to now.
        auth: Auth, optional
            Authorization object used to authenticate the REST API.
            If provided it will be prioritized over global authentication.
        request_timeout: int, optional
            Seconds before giving up a request without an answer.
        request_retries: int, optional
            Maximum number of times to retry a request before giving up.

        Returns
        -------
        history : EventHistory
            Object containing each event in history in addition to a few
            convenience functions.

        """

        # Construct URL.
        url = '/projects/{}/devices/{}/events'.format(project_id, device_id)

        # Construct parameters dictionary.
        params: dict = dict()
        if event_types is not None:
            params['eventTypes'] = event_types

        # Sanitize timestamps as they must be iso8601 format.
        start_time_iso8601 = dttrans.to_iso8601(start_time)
        if start_time_iso8601 is not None:
            params['startTime'] = start_time_iso8601
        end_time_iso8601 = dttrans.to_iso8601(end_time)
        if end_time_iso8601 is not None:
            params['endTime'] = end_time_iso8601

        # Send paginated GET request.
        res = dtrequests.DTRequest.paginated_get(
            url=url,
            pagination_key='events',
            params=params,
            **kwargs,
        )

        # Return list of Event objects of paginated GET response.
        return cls(Event.from_mixed_list(res))

    def get_events(self, event_type):
        """
        Returns a filtered list of events specified by type from history.

        Parameters
        ----------
        event_type : str
            The event type to be returned.

        Returns
        -------
        events : list
            List of event objects of the specified type.

        """

        out = []
        for e in self.events_list:
            if e.event_type == event_type:
                out.append(e)
        return out

    def get_data_axes(self,
                      x_name: str,
                      y_name: Optional[str] = None,
                      event_type: Optional[str] = None,
                      ):
        """
        Returns filtered lists of the data axes as specified from
        event type in history.

        Parameters
        ----------
        x_name : str
            Name of the first axis to isolate.
        y_name : str, optional
            Name of the second axis to isolate.
        event_type : str, optional
            Filter by event type. For instance, "state" exists for both
            waterPresent and objectPresent event types.

        Returns
        -------
        x_axis : list
            List of values on the first axis.
        y_axis : list
            List of values on the second axis.

        """

        x_axis = []
        y_axis = []
        for e in self.events_list:
            # Skip event if event_type mismatch.
            if event_type is not None and event_type != e.event_type:
                continue

            # Skip event if x- or y-name does not exist.
            if not hasattr(e.data, x_name):
                continue
            if y_name is not None and not hasattr(e.data, y_name):
                continue

            # Tests passed, append timestamp and y-axis.
            x_axis.append(getattr(e.data, x_name))
            if y_name is not None:
                y_axis.append(getattr(e.data, y_name))

        if y_name is not None:
            return x_axis, y_axis
        else:
            return x_axis



# Standard library imports.
import re
import base64
from datetime import datetime

# Project imports.
import disruptive.errors as dterrors


def base64_encode(string: str):
    string_bytes = string.encode('ascii')
    base64_bytes = base64.b64encode(string_bytes)
    base64_string = base64_bytes.decode('ascii')
    return base64_string


def to_iso8601(ts):
    # Verify that we even got a string.
    if isinstance(ts, str):
        # As it is a string, verify iso8601 format.
        if validate_iso8601_format(ts):
            # Nothing to do, return as is.
            return ts
        else:
            # Invalid iso8601 format, raise error.
            raise dterrors.FormatError(
                'Timestamp format {} is invalid iso8601 format.\n'
                'E.g. 2020-01-01T00:00:00Z'.format(
                    ts
                )
            )

    # If not string, datetime is also fine as it can be converted.
    elif isinstance(ts, datetime):
        # Use built-in functions for converting to string.
        return ts.isoformat() + 'Z'

    # If ts is None, return None.
    elif ts is None:
        return None

    # If any other type, raise TypeError.
    else:
        raise TypeError(
            'Got timestamp of type {}, expected'
            ' iso8601 str or datetime.'.format(
                type(ts).__name__
            )
        )


def to_datetime(ts):
    # Check if input is already datetime format.
    if isinstance(ts, datetime):
        # Nothing to do, just return it.
        return ts

    # If input is string, we might be able to convert it.
    elif isinstance(ts, str):
        # First, verify if string is valid iso8601 format.
        if validate_iso8601_format(ts):
            # Use built-in functions for converting to datetime.
            return datetime.fromisoformat(ts.replace('Z', '+00:00'))
        else:
            # Invalid iso8601 format, raise error.
            raise dterrors.FormatError(
                'Timestamp format {} is invalid iso8601 format.\n'
                'E.g. 2020-01-01T00:00:00Z'
            )

    # If ts is None, return None.
    elif ts is None:
        return None

    # If any other type, raise TypeError.
    else:
        raise TypeError(
            'Got timestamp of type {}, expected'
            ' iso8601 str or datetime.'.format(
                type(ts).__name__
            )
        )


def validate_iso8601_format(dt_str):
    # Set up regex for matching iso8601 string.
    # This should probably be changed in the future as it is
    # a little forced. However, the reason for using this approach is
    # that the datetime built-in method for checking iso8601 format
    # allows missing timezone infromation (i.e. Z or +-00:00 suffix).
    # This must be included in our API, and is why this regex exists.
    iso8601_regex = r'^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-' \
        '(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):([0-5][0-9]):' \
        '([0-5][0-9])?(.[0-9]+)?(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])$'
    match_iso8601 = re.compile(iso8601_regex).match

    if match_iso8601(dt_str) is not None:
        return True
    else:
        return False
